# FilterEditor.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Editor component for generic filter ASTs.
# Copyright 2008 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""Widget for ediging filters."""

from sutekh.gui.MultiSelectComboBox import MultiSelectComboBox
from sutekh.gui.SutekhDialog import SutekhDialog, do_complaint_error
from sutekh.gui.AutoScrolledWindow import AutoScrolledWindow
from sutekh.core.FilterParser import FilterNode, BinOpNode, NotOpNode, \
        FilterPartNode
from sutekh.core import FilterParser
import gtk
import pango
import gobject

class FilterEditor(gtk.Frame):
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    """GTK component for editing Sutekh filter ASTs.

       Provides a graphical representation of the filter as nested boxes,
       which the user can extend or remove from.
       """
    def __init__(self, oAST, sFilterType, oParser, oFilterDialog):
        super(FilterEditor, self).__init__(" Filter Editor ")
        self.__sFilterType = sFilterType
        self.__oParser = oParser
        self.__oFilterDialog = oFilterDialog # Dialog we're a child of

        self.__oBoxModel = None
        self.__oBoxEditor = None

        oNameLabel = gtk.Label("Filter name:")
        self.__oNameEntry = gtk.Entry()
        self.__oNameEntry.set_width_chars(30)

        oTextEditorButton = gtk.Button("Edit Query Text")
        oTextEditorButton.connect("clicked", self.__show_text_editor)
        oHelpButton = gtk.Button("Help")
        oHelpButton.connect("clicked", self.__show_help_dialog)

        oHBox = gtk.HBox(spacing=5)
        oHBox.pack_start(oNameLabel, expand=False)
        oHBox.pack_start(self.__oNameEntry, expand=False)
        oHBox.pack_end(oHelpButton, expand=False)
        oHBox.pack_end(oTextEditorButton, expand=False)

        self.__oVBox = gtk.VBox(spacing=5)
        self.__oVBox.pack_end(oHBox, expand=False)
        self.__oVBox.pack_end(gtk.HSeparator(), expand=False)

        self.add(self.__oVBox)

        self.replace_ast(oAST)

    def get_filter(self):
        """Get the actual filter for the current AST."""
        oAST = self.get_current_ast()
        if oAST is None:
            return None
        else:
            return oAST.get_filter()

    def get_current_ast(self):
        """Get the current AST represented by the editor."""
        oNewAST = self.__oBoxModel.get_ast()
        if oNewAST is None:
            return oNewAST

        dValues = self.get_current_values()
        for aVals in dValues.values():
            if not aVals:
                do_complaint_error("Some filter values have not been " \
                                   "filled in so the selected filter " \
                                   "will be ignored")
                return None

        aNewValues = oNewAST.get_values()
        for oValue in aNewValues:
            if oValue.is_entry() or oValue.is_list() or oValue.is_tuple():
                oValue.oNode.set_values(dValues[oValue.oNode.get_name()])

        return oNewAST

    def replace_ast(self, oAST):
        """Replace the current AST with a new one and update the GUI,
           preserving variable values if possible.

           Also used to setup the filter initially.
           """
        if self.__oBoxEditor is None:
            dVars = {}
        else:
            dVars = self.get_current_values()
            self.__oVBox.remove(self.__oBoxEditor)

        self.__oBoxModel = FilterBoxModel(oAST, self.__sFilterType)
        self.__oBoxEditor = FilterBoxModelEditor(self.__oBoxModel,
                self.__oFilterDialog)
        self.__oVBox.pack_start(self.__oBoxEditor)
        self.show_all()

        self.set_current_values(dVars)

    def get_current_values(self):
        """Get the current values in the editor."""
        return self.__oBoxEditor.get_current_values()

    def set_current_values(self, dVars):
        """Set the values in the editor to those in dVars."""
        self.__oBoxEditor.set_current_values(dVars)

    def get_current_text(self):
        """Get the current text in the editor."""
        return self.__oBoxModel.get_text()

    def set_name(self, sName):
        """Set the filter name."""
        self.__oNameEntry.set_text(sName)

    def get_name(self):
        """Get the filter name."""
        return self.__oNameEntry.get_text().strip()

    def connect_name_changed(self, fCallback):
        """Connect a callback to the name entry change signal."""
        self.__oNameEntry.connect('changed', fCallback)

    # pylint: disable-msg=W0613
    # oTextEditorButton required by function signature
    def __show_text_editor(self, oTextEditorButton):
        """Show a gtk.Entry widget so filters can be directly typed by the
           user."""
        oDlg = SutekhDialog("Query Editor", self.__oFilterDialog,
                gtk.DIALOG_DESTROY_WITH_PARENT,
                (gtk.STOCK_OK, gtk.RESPONSE_OK,
                 gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))

        oDlg.set_default_size(600, -1)

        oEntry = gtk.Entry()
        oEntry.set_text(self.get_current_text())
        # pylint: disable-msg=E1101
        # vbox confuse pylint
        oDlg.vbox.pack_start(oEntry)
        oDlg.show_all()

        try:
            iResponse = oDlg.run()
            if iResponse == gtk.RESPONSE_OK:
                sNewFilter = oEntry.get_text()
                try:
                    oAST = self.__oParser.apply(sNewFilter)
                except ValueError, oExcep:
                    do_complaint_error("Invalid Filter: %s\n Error: %s"
                            % (sNewFilter, str(oExcep)))
                else:
                    self.replace_ast(oAST)
        finally:
            oDlg.destroy()

    # oHelpButton required by function signature
    def __show_help_dialog(self, oHelpButton):
        """Show a dialog window with the helptext from the filters."""
        oDlg = SutekhDialog("Help on Filters", self.__oFilterDialog,
                gtk.DIALOG_DESTROY_WITH_PARENT,
                (gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE))

        oDlg.set_default_size(600, 500)
        # pylint: disable-msg=E1101
        # vbox confuse pylint

        oHelpView = AutoScrolledWindow(FilterHelpTextView(
            self.__oBoxModel.get_filter_types()))
        oDlg.vbox.pack_start(oHelpView)
        oDlg.show_all()

        try:
            oDlg.run()
        finally:
            oDlg.destroy()


class FilterBoxModel(list):
    """Converts a filter AST into a simple nest box model.

       Each box either ANDs or ORs all of its contained filters.
       Each contained filter may optionally be negated.
       Values are retained for each contained filter.
       """

    AND, OR = 'and', 'or'

    # pylint: disable-msg=W0231
    # no point to calling list's __init__
    def __init__(self, oAST, sFilterType, oVarNameMaker=None):
        """Initialise a filter box from an AST filter representation."""
        if oVarNameMaker is None:
            oVarNameMaker = VariableNameGenerator()
            bVarMakerNeedsInit = True
        else:
            bVarMakerNeedsInit = False

        self.sFilterType = sFilterType
        self.bNegate = False
        self.oVarNameMaker = oVarNameMaker

        if type(oAST) is FilterNode:
            oAST = oAST.oExpression

        if type(oAST) is BinOpNode:
            self.sBoxType = oAST.oOp
            self._init_binop(oAST)
        elif type(oAST) is NotOpNode and \
                type(oAST.oSubExpression) is BinOpNode:
            self.bNegate = True
            self.sBoxType = oAST.oSubExpression.oOp
            self._init_binop(oAST.oSubExpression)
        elif type(oAST) is FilterPartNode or \
                type(oAST) is NotOpNode and \
                type(oAST.oSubExpression) is FilterPartNode:
            self.sBoxType = self.AND
            self.append(FilterBoxItem(oAST))
        elif oAST is None:
            # support for completely empty boxes
            self.sBoxType = self.AND
        else:
            raise ValueError("FilterBoxModel cannot represent AST %s of"
                    " type %s" % (oAST, type(oAST)))

        assert self.sBoxType in [self.AND, self.OR]

        if bVarMakerNeedsInit:
            self.oVarNameMaker.update(self.get_variable_names())

    def _init_binop(self, oBinOp):
        """Create the correct box entries for a BinOpNode in the AST."""
        for oChild in [oBinOp.oLeft, oBinOp.oRight]:
            if type(oChild) is BinOpNode and oChild.oOp == oBinOp.oOp:
                self._init_binop(oChild)
            elif type(oChild) is BinOpNode:
                self.append(FilterBoxModel(oChild, self.sFilterType,
                    self.oVarNameMaker))
            elif type(oChild) in [NotOpNode, FilterPartNode]:
                self.append(FilterBoxItem(oChild))
            else:
                raise ValueError("FilterBoxModel encountered unsupported AST"
                        " node type %s (%s) while examing BinOpNode tree." %
                        (type(oChild), oChild))

    def set_boxtype(self, sBoxType, bNegate=False):
        """Set the type for this filter box"""
        self.sBoxType = sBoxType
        self.bNegate = bNegate

    def get_variable_names(self):
        """Get the variable names of the children in this box."""
        oNames = set()
        for oChild in self:
            oNames.update(oChild.get_variable_names())
        return oNames

    def get_ast(self):
        """Return an AST representation of the filter."""
        aChildASTs = [oChild.get_ast() for oChild in self]
        aChildASTs = [oChild for oChild in aChildASTs if oChild is not None]

        if len(aChildASTs) == 0:
            oAST = None
        elif len(aChildASTs) == 1:
            oAST = aChildASTs[0]
        else:
            oAST = BinOpNode(aChildASTs[0], self.sBoxType, aChildASTs[1])
            for oChild in aChildASTs[2:]:
                oAST = BinOpNode(oAST, self.sBoxType, oChild)

        if self.bNegate and oAST is not None:
            oAST = NotOpNode(oAST)

        return oAST

    def get_text(self):
        """Return a text representation of the filter."""
        aChildTexts = [oChild.get_text() for oChild in self]
        aChildTexts = [sChild for sChild in aChildTexts if sChild]

        if len(aChildTexts) == 0:
            sText = ''
        elif len(aChildTexts) == 1:
            sText = aChildTexts[0]
        else:
            sText = "(" + (") %s (" % self.sBoxType).join(aChildTexts) + ")"

        if self.bNegate and sText:
            sText = "NOT (%s)" % (sText,)

        return sText

    def get_filter_types(self):
        """Get the types support by this filter."""
        return [oFilterType for oFilterType in FilterParser.aParserFilters
                if self.sFilterType in oFilterType.types]

    def add_child_box(self, sChildBoxType):
        """Add a child box to this box."""
        assert sChildBoxType in [self.AND, self.OR]
        oChildBox = FilterBoxModel(None, self.sFilterType, self.oVarNameMaker)
        oChildBox.sBoxType = sChildBoxType
        self.append(oChildBox)
        return oChildBox

    def add_child_item(self, sChildTypeKeyword):
        """Add a filter item to this box."""
        sVarName = self.oVarNameMaker.generate_name()
        oAST = FilterPartNode(sChildTypeKeyword, None, sVarName)
        oChildItem = FilterBoxItem(oAST)
        self.append(oChildItem)
        return oChildItem

    def remove_child(self, oChild):
        """Remove a filter item or box from this box."""
        self.remove(oChild)

class FilterBoxModelEditor(gtk.VBox):
    """Widget for editing a FilterBoxModel."""
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    BOXTYPE = { # description -> (AND or OR, bNegate)
        'All of ...': (FilterBoxModel.AND, False),
        'Any of ...': (FilterBoxModel.OR, False),
        'Not all of ...': (FilterBoxModel.AND, True),
        'None of ...': (FilterBoxModel.OR, True),
    }

    BOXTYPE_ORDER = [ # order types should appear in combo
        'All of ...', 'Any of ...', 'Not all of ...',
        'None of ...',
    ]

    def __init__(self, oBoxModel, oParentWin):
        super(FilterBoxModelEditor, self).__init__(spacing=5)
        self.__oBoxModel = oBoxModel
        self.__aChildren = []
        self.__fRemoveBox = None # set by parent (if any)
        self.__oParentWin = oParentWin

        self.__oBoxTypeSelector = gtk.combo_box_new_text()
        self.pack_start(self.__oBoxTypeSelector, expand=False)

        for iLoop, sDesc in enumerate(self.BOXTYPE_ORDER):
            self.__oBoxTypeSelector.append_text(sDesc)
            sBoxType, bNegate = self.BOXTYPE[sDesc]
            if self.__oBoxModel.sBoxType == sBoxType \
               and self.__oBoxModel.bNegate == bNegate:
                self.__oBoxTypeSelector.set_active(iLoop)

        self.__oBoxTypeSelector.connect('changed', self.__change_boxtype)

        oChildHBox = gtk.HBox(spacing=10)
        oChildHBox.pack_start(gtk.VSeparator(), expand=False)
        self.__oChildArea = gtk.VBox(spacing=5)
        oChildHBox.pack_start(self.__oChildArea)
        self.pack_start(oChildHBox, expand=False)

        for oChild in self.__oBoxModel:
            if type(oChild) is FilterBoxModel:
                oModelEditor = FilterBoxModelEditor(oChild, self.__oParentWin)
                oModelEditor.register_remove_box(self.__remove_model)
                self.__aChildren.append(oModelEditor)
                self.__oChildArea.pack_start(oModelEditor, expand=False)
            else:
                oItemEditor = FilterBoxItemEditor(oChild, self.__oParentWin)
                oItemEditor.connect_remove_button(self.__remove_filter_part)
                self.__aChildren.append(oItemEditor)
                self.__oChildArea.pack_start(oItemEditor, expand=False)

        self.pack_start(gtk.HSeparator(), expand=False)

        oHBox = gtk.HBox()
        self.pack_start(oHBox, expand=False)

        oAddButton = gtk.Button("+")
        oHBox.pack_start(oAddButton, expand=False)

        oTypeSelector = self._make_type_list()
        oHBox.pack_start(oTypeSelector, expand=False)

        oAddButton.connect('clicked', self.__add_filter_part, oTypeSelector)

    def _make_type_list(self):
        """Create a combo box to select the filter type."""
        oTypeStore = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_STRING,)
        oTypeSelector = gtk.ComboBox(oTypeStore)
        oCell = gtk.CellRendererText()
        oTypeSelector.pack_start(oCell, True)
        oTypeSelector.add_attribute(oCell, 'text', 0)

        for oFilterType in self.__oBoxModel.get_filter_types():
            oIter = oTypeStore.append(None)
            oTypeStore.set(oIter, 0, oFilterType.description,
                                  1, oFilterType.keyword)
        oIter = oTypeStore.append(None)
        oTypeStore.set(oIter, 0, "Sub-Filter",
                              1, None)

        return oTypeSelector

    def get_title(self):
        """Get the correct title for this box"""
        if self.__oBoxModel.sBoxType == FilterBoxModel.AND:
            return "All of ..."
        else:
            return "One of ..."

    def get_current_values(self):
        """Return a dictionary of name -> value mappings for the variables
           present in the filter editing gui.
           """
        dVars = {}
        for oChild in self.__aChildren:
            dVars.update(oChild.get_current_values())
        return dVars

    def set_current_values(self, dVars):
        """Set the current list values using variable names in the given
           dictionary.

           dVars is a mapping sVariableName -> [ list of string values ]
           """
        for oChild in self.__aChildren:
            oChild.set_current_values(dVars)

    def register_remove_box(self, fRemoveBox):
        """Register a function for removing this box from it's parent editor.

           The function signature is fRemoveBox(oBoxEditor, oBoxModel)
           """
        self.__fRemoveBox = fRemoveBox
        self.__oBoxTypeSelector.append_text("Remove Sub-Filter")
        self.__oBoxTypeSelector.show_all()

    def __change_boxtype(self, oBoxTypeSelector):
        """Change the type of this box to that chosen by the user."""
        sType = oBoxTypeSelector.get_active_text()
        if sType == "Remove Sub-Filter":
            self.__fRemoveBox(self, self.__oBoxModel)
        else:
            sBoxType, bNegate = self.BOXTYPE[sType]
            self.__oBoxModel.set_boxtype(sBoxType, bNegate)

    # pylint: disable-msg=W0613
    # oAddButton needed by function signature
    def __add_filter_part(self, oAddButton, oTypeSelector):
        """Add a filter part or filter box to this box."""
        oIter = oTypeSelector.get_active_iter()
        oModel = oTypeSelector.get_model()

        if not oIter:
            # ignore add button clicks when a type isn't selected
            return

        sDescription = oModel.get_value(oIter, 0)
        sKeyword = oModel.get_value(oIter, 1)

        if sDescription == "Sub-Filter":
            oChildBoxModel = self.__oBoxModel.add_child_box(FilterBoxModel.AND)
            oChildEditor = FilterBoxModelEditor(oChildBoxModel,
                    self.__oParentWin)
            oChildEditor.register_remove_box(self.__remove_model)
        else:
            oChildItem = self.__oBoxModel.add_child_item(sKeyword)
            oChildEditor = FilterBoxItemEditor(oChildItem, self.__oParentWin)
            oChildEditor.connect_remove_button(self.__remove_filter_part)

        self.__aChildren.append(oChildEditor)
        self.__oChildArea.pack_start(oChildEditor, expand=False)
        self.show_all()

    # oRemoveButton needed by function signature
    def __remove_filter_part(self, oRemoveButton, oEditor, oModelOrItem):
        """Remove the filter part from this box at the user's request"""
        self.__oBoxModel.remove_child(oModelOrItem)
        self.__aChildren.remove(oEditor)
        self.__oChildArea.remove(oEditor)

    def __remove_model(self, oEditor, oModel):
        """Remove the filter model from this box"""
        self.__remove_filter_part(None, oEditor, oModel)

class FilterBoxItem(object):
    """A item in the filter editor.

       This represents either a single FilterPart or a single
       NOT(FilterPart) expression in the AST.
       """
    NONE, ENTRY, LIST, LIST_FROM = range(4)

    def __init__(self, oAST):
        if type(oAST) is NotOpNode:
            self.bNegated = True
            oAST = oAST.oSubExpression
        else:
            self.bNegated = False

        assert type(oAST) is FilterPartNode

        self.sFilterName = oAST.sFilterName
        self.sVariableName = oAST.sVariableName
        self.aFilterValues = oAST.aFilterValues
        self.iValueType = None

        # process values
        self.sLabel, self.aValues = None, None
        for oValue in oAST.get_values():
            if oValue.is_value():
                assert self.sLabel is None
                self.sLabel = oValue.oValue
            elif oValue.is_list():
                assert self.iValueType is None
                self.aValues = oValue.oValue
                self.iValueType = self.LIST
            elif oValue.is_tuple():
                assert self.iValueType is None
                self.aValues = oValue.oValue
                self.iValueType = self.LIST_FROM
            elif oValue.is_entry():
                assert self.iValueType is None
                self.iValueType = self.ENTRY
            elif oValue.is_None():
                assert self.iValueType is None
                self.iValueType = self.NONE

        assert self.sLabel is not None
        assert self.iValueType is not None

    def get_variable_names(self):
        """Get the variable name for this filter"""
        return set([self.sVariableName])

    def get_ast(self):
        """Return an AST respresentation of the filter."""
        oAST = FilterPartNode(self.sFilterName, self.aFilterValues,
                self.sVariableName)
        if self.bNegated:
            oAST = NotOpNode(oAST)
        return oAST

    def get_text(self):
        """Return a text representation of the filter."""
        if self.iValueType == self.NONE:
            sText = self.sFilterName
        else:
            sText = "%s in %s" % (self.sFilterName, self.sVariableName)
        return sText

class FilterBoxItemEditor(gtk.HBox):
    """Widget for editing the entries for the filter associated with a
       FilterBoxItem."""
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    def __init__(self, oBoxItem, oParent):
        super(FilterBoxItemEditor, self).__init__(spacing=5)
        self.__oBoxItem = oBoxItem

        self.__oRemoveButton = gtk.Button("-")
        self.pack_start(self.__oRemoveButton, expand=False)

        self.__oNegateButton = gtk.CheckButton("NOT")
        self.__oNegateButton.set_active(self.__oBoxItem.bNegated)
        self.__oNegateButton.connect('toggled', self.__toggle_negated)
        self.pack_start(self.__oNegateButton, expand=False)

        self.pack_start(gtk.Label(self.__oBoxItem.sLabel), expand=False)

        if self.__oBoxItem.iValueType == FilterBoxItem.LIST:
            oWidget = MultiSelectComboBox(oParent)
            oWidget.fill_list(self.__oBoxItem.aValues)
            oWidget.set_list_size(200, 400)
        elif self.__oBoxItem.iValueType == FilterBoxItem.LIST_FROM:
            # Create an additional widget for the from side
            oWidget = MultiSelectComboBox(oParent)
            oWidget.fill_list(self.__oBoxItem.aValues[0])
            oWidget.set_list_size(200, 400)
            oFromStore = gtk.ListStore(str)
            self.__oFromWidget = gtk.ComboBox(oFromStore)
            oCell = gtk.CellRendererText()
            self.__oFromWidget.pack_start(oCell, True)
            self.__oFromWidget.add_attribute(oCell, 'text', 0)
            for sCardSet in self.__oBoxItem.aValues[1]:
                oIter = oFromStore.append(None)
                oFromStore.set(oIter, 0, sCardSet)
            self.pack_end(self.__oFromWidget, False, False)
            self.pack_end(gtk.Label(' from '), False, False)
        elif self.__oBoxItem.iValueType == FilterBoxItem.NONE:
            oWidget = gtk.Label('  < No user data required > ')
        elif self.__oBoxItem.iValueType == FilterBoxItem.ENTRY:
            oWidget = gtk.Entry(100)
            oWidget.set_width_chars(30)
        else:
            raise RuntimeError("Unknown FilterBoxItem ValueType %s" %
                    (self.__oBoxItem.iValueType,))

        self.__oEntryWidget = oWidget
        self.pack_start(oWidget)

    def connect_remove_button(self, fHandler):
        """Add a handler for the 'remove filter' button"""
        if self.__oRemoveButton is not None:
            self.__oRemoveButton.connect('clicked', fHandler, self,
                    self.__oBoxItem)

    def get_current_values(self):
        """Get the current values set for this filter."""
        dVars = {}
        sName = self.__oBoxItem.sVariableName
        if self.__oBoxItem.aValues:
            aSelection = self.__oEntryWidget.get_selection()
            if self.__oBoxItem.sFilterName in FilterParser.aWithFilters:
                aSplit = [sItem.split(" with ") for sItem in aSelection]
                aVals = ['"%s with %s"' % (sPart1, sPart2) for sPart1,
                        sPart2 in aSplit]
            elif self.__oBoxItem.sFilterName in FilterParser.aFromFilters:
                # Need to sort out how to handle this case
                oIter = self.__oFromWidget.get_active_iter()
                oModel = self.__oFromWidget.get_model()
                if not oIter:
                    sFrom = None
                else:
                    sFrom = '"%s"' % oModel.get_value(oIter, 0)
                aVals = [['"%s"' % (sItem,) for sItem in aSelection], sFrom]
            else:
                aVals = ['"%s"' % (sItem,) for sItem in aSelection]
            dVars[sName] = aVals
        else:
            sText = self.__oEntryWidget.get_text()
            if sText != '':
                dVars[sName] = ['"' + sText + '"']
        return dVars

    def set_current_values(self, dVars):
        """Set the current values for the entry widget for this filter
           item."""
        sName = self.__oBoxItem.sVariableName
        if not sName in dVars:
            return
        if self.__oBoxItem.aValues:
            if self.__oBoxItem.iValueType == FilterBoxItem.LIST:
                aVals = [sVal.strip('"') for sVal in dVars[sName]]
                self.__oEntryWidget.set_selection(aVals)
            elif self.__oBoxItem.iValueType == FilterBoxItem.LIST_FROM:
                if len(dVars[sName]) != 2:
                    return
                aVals = [sVal.strip('"') for sVal in dVars[sName][0]]
                self.__oEntryWidget.set_selection(aVals)
                sFrom = dVars[sName][1]
                if sFrom:
                    sFrom = sFrom.strip('"')
                    oModel = self.__oFromWidget.get_model()
                    oIter = oModel.get_iter_first()
                    while oIter:
                        sEntry = oModel.get(oIter, 0)[0]
                        if sEntry == sFrom:
                            self.__oFromWidget.set_active_iter(oIter)
                            break
                        oIter = oModel.iter_next(oIter)
        elif len(dVars[sName]) > 0:
            self.__oEntryWidget.set_text(dVars[sName][0].strip('"'))
        else:
            self.__oEntryWidget.set_text("")

    # pylint: disable-msg=W0613
    # oWidget is needed by function signature
    def __toggle_negated(self, oWidget):
        """Response to the use setting/unsetting the not status of this
           filter item"""
        self.__oBoxItem.bNegated = self.__oNegateButton.get_active()

class VariableNameGenerator(set):
    """Generate a unique name for a variable in a filter."""
    KEYFORM = "$var%s"

    def __init__(self):
        super(VariableNameGenerator, self).__init__()
        self.__iNum = 0

    def generate_name(self):
        """Generate the next variable in the list."""
        sName = self.KEYFORM % self.__iNum
        while sName in self:
            self.__iNum += 1
            sName = self.KEYFORM % self.__iNum
        self.add(sName)
        return sName

class FilterHelpBuffer(gtk.TextBuffer, object):
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    """TextBuffer object used to display the help about the filters."""
    def __init__(self):
        super(FilterHelpBuffer, self).__init__(None)

        # See http://www.pygtk.org/pygtk2reference/class-gtktexttag.html
        # for some possible properties

        self.create_tag("description", weight=pango.WEIGHT_BOLD)
        self.create_tag("keyword", style=pango.STYLE_ITALIC)
        self.create_tag("helptext", left_margin=10)
        self._oIter = self.get_iter_at_offset(0)

    # pylint: disable-msg=W0142
    # ** magic OK here
    def tag_text(self, *aArgs, **kwargs):
        """Insert text into the buffer with tags."""
        self.insert_with_tags_by_name(self._oIter, *aArgs, **kwargs)

    # pylint: enable-msg=W0142

    def add_help_text(self, sDescription, sKeyword, sHelpText):
        """Add the given help text to the text buffer."""
        self.tag_text(sDescription, "description")
        self.tag_text(" (")
        self.tag_text(sKeyword, "keyword")
        self.tag_text(") :\n")
        self.tag_text(sHelpText, "helptext")
        self.tag_text("\n")

    def reset_iter(self):
        """reset the iterator to the start of the buffer."""
        self._oIter = self.get_iter_at_offset(0)

class FilterHelpTextView(gtk.TextView, object):
    """TextView widget for displaying the help text."""
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    def __init__(self, aFilterTypes):
        super(FilterHelpTextView, self).__init__()
        oBuf = FilterHelpBuffer()

        self.set_buffer(oBuf)
        self.set_editable(False)
        self.set_cursor_visible(False)
        self.set_wrap_mode(gtk.WRAP_WORD)
        self.set_border_width(5)
        self.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse("white"))

        oBuf.reset_iter()

        oBuf.tag_text(
            "The available filtering options are listed below. "
            "The first line of each item shows the description "
            "you'll see in the filter editor in bold with the "
            "keyword to use in query strings in italics afterwards. "
            "The rest of the description describes the arguments "
            "the filter takes and the results it produces."
            "\n\n"
        )

        for oFilt in aFilterTypes:
            oBuf.add_help_text(oFilt.description, oFilt.keyword,
                    oFilt.helptext)

        oBuf.tag_text(
            "\n"
            "Note: Using filters which require information about "
            "cards in a card set will temporarily turn off displaying "
            "cards with zero counts (for modes where this is possible) "
            "in the card list since there are no "
            "associated cards on which to base the query. "
            "Filters affected by this are: Physical Expansion, "
            "Card Sets and In Card Sets in Use."
        )



# ScrolledList.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2006 Neil Muller <drnlmuller+sutekh@gmail.com>
# Copyright 2006 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""Generic Scrolled List, used in the Filter Dialo and elsewhere"""

import gtk, gobject
from sutekh.gui.AutoScrolledWindow import AutoScrolledWindow

class ScrolledList(gtk.Frame):
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    """Frame containing an auto scrolled list"""
    def __init__(self, sTitle):
        super(ScrolledList, self).__init__(None)
        self._oListStore = gtk.ListStore(gobject.TYPE_STRING)
        self._oTreeView = gtk.TreeView(self._oListStore)
        oCell1 = gtk.CellRendererText()
        oColumn1 = gtk.TreeViewColumn(sTitle, oCell1, markup=0)
        self.dSecondColVals = {}
        self._oTreeView.append_column(oColumn1)
        self._oTreeView.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
        oMyScroll = AutoScrolledWindow(self._oTreeView)
        self.add(oMyScroll)
        self.set_shadow_type(gtk.SHADOW_NONE)
        self.show_all()

    # pylint: disable-msg=W0212
    # allow access via these properties
    view = property(fget=lambda self: self._oTreeView,
            doc="Associated View Object")
    itemlist = property(fget=lambda self: self._oListStore,
            doc="List of values")
    # disable-msg=W0212

    def add_second_column(self, sTitle):
        """Add an extra column to the list."""
        oCell2 = gtk.CellRendererText()
        oColumn2 = gtk.TreeViewColumn(sTitle, oCell2)
        self._oTreeView.append_column(oColumn2)
        oColumn2.set_cell_data_func(oCell2, self._render_second_column)

    def set_select_single(self):
        """set selection to single mode"""
        self._oTreeView.get_selection().set_mode(gtk.SELECTION_SINGLE)

    def set_select_multiple(self):
        """set selection to multiple mode"""
        self._oTreeView.get_selection().set_mode(gtk.SELECTION_MULTIPLE)

    def set_select_none(self):
        """set selection mode to none"""
        self._oTreeView.get_selection().set_mode(gtk.SELECTION_NONE)

    def get_selection(self):
        """Return a list of the selected elements of the list"""
        aSelectedList = []
        oModel, oSelection = \
                self._oTreeView.get_selection().get_selected_rows()
        for oPath in oSelection:
            oIter = oModel.get_iter(oPath)
            sName = oModel.get_value(oIter, 0)
            aSelectedList.append(sName)
        return aSelectedList

    def set_selection(self, aRowsToSelect):
        """Set the selected rows to aRowsToSelect."""
        aRowsToSelect = set(aRowsToSelect)
        oIter = self._oListStore.get_iter_first()
        oTreeSelection = self._oTreeView.get_selection()
        oTreeSelection.unselect_all()
        while oIter is not None:
            sName = self._oListStore.get_value(oIter, 0)
            if sName in aRowsToSelect:
                oTreeSelection.select_iter(oIter)
            oIter = self._oListStore.iter_next(oIter)

    def set_selected(self, sEntry):
        """Select the current entry"""
        oIter = self._oListStore.get_iter_first()
        while oIter is not None:
            sName = self._oListStore.get_value(oIter, 0)
            if sName == sEntry:
                oPath = self._oListStore.get_path(oIter)
                self._oTreeView.set_cursor(oPath)
                return
            oIter = self._oListStore.iter_next(oIter)

    def fill_list(self, aVals):
        """Fill the list store with the given values"""
        self._oListStore.clear()
        for sEntry in aVals:
            oIter = self._oListStore.append(None)
            self._oListStore.set(oIter, 0, sEntry)

    def fill_second_column(self, dVals):
        """Fill the values in the second column."""
        self.dSecondColVals = dVals
        self._oTreeView.queue_draw()

    # pylint: disable-msg=W0613
    # oColumn required by function signature
    def _render_second_column(self, oColumn, oCell, oModel, oIter):
        """Method to render the values in the second column."""
        sKey = oModel.get_value(oIter, 0)
        if sKey in self.dSecondColVals:
            oCell.set_property("markup", self.dSecondColVals[sKey])
        else:
            oCell.set_property("markup", "")






# PluginManager.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2006 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""Classes for mangaging and creating plugins for Sutekh."""

import os
import glob
import logging
import sutekh.gui.plugins as plugins
from sutekh.core.DatabaseVersion import DatabaseVersion
from sutekh.core.SutekhObjects import PhysicalCardSet

class PluginManager(object):
    """Manages plugins for Sutekh

       Plugin modules should be placed in the plugins package directory and
       contain an attribute named 'plugin' which points to the plugin class the
       module contains.
       """

    def __init__(self):
        self._aCardListPlugins = []

    def load_plugins(self, bVerbose):
        """Load list of Plugin Classes from plugin dir."""
        sPluginDir = os.path.dirname(plugins.__file__)

        for sPluginPath in glob.glob(os.path.join(sPluginDir, "*.py")):
            sPluginName = os.path.basename(sPluginPath)[:-len(".py")]

            if sPluginName == "__init__":
                continue

            # load module
            # pylint: disable-msg=C0103
            # mPlugin is legal name here
            try:
                mPlugin = __import__("sutekh.gui.plugins.%s" % sPluginName,
                        None, None, [plugins])
            except ImportError, oExp:
                if bVerbose:
                    logging.warn("Failed to load plugin %s (%s)." % (
                        sPluginName, oExp))
                continue

            # find plugin class
            try:
                cPlugin = mPlugin.plugin
            except AttributeError, oExp:
                if bVerbose:
                    logging.warn("Plugin module %s appears not to contain a"
                            " plugin (%s)." % (sPluginName, oExp))
                continue

            # add to appropriate plugin lists
            if issubclass(cPlugin, CardListPlugin):
                self._aCardListPlugins.append(cPlugin)

    def get_card_list_plugins(self):
        """Get all the plugins loaded"""
        return list(self._aCardListPlugins)

class CardListPlugin(object):
    """Base class for card list plugins."""
    dTableVersions = {}
    aModelsSupported = []

    def __init__(self, oCardListView, oCardListModel, cModelType=None):
        """oCardListModel - card list model for this plugin to operate on."""
        self._oView = oCardListView
        self._oModel = oCardListModel
        self._cModelType = cModelType

    # pylint: disable-msg=W0212
    # we allow access to the members via these properties
    parent = property(fget=lambda self: self._oView.mainwindow,
            doc="Parent window to use when creating dialogs.")
    view = property(fget=lambda self: self._oView,
            doc="Associated CardListView object.")
    model = property(fget=lambda self: self._oModel,
            doc="Associated CardModel object.")
    cardlookup = property(fget=lambda self: self.parent.cardLookup,
            doc="GUI CardLookup.")
    icon_manager = property(fget=lambda self: self.parent.icon_manager,
            doc="Icon manager.")
    # pylint: enable-msg=W0212

    def add_to_menu(self, dAllMenus, oCatchAllMenu):
        """Grunt work of adding menu item to the frame"""
        aMenuItems = self.get_menu_item()
        if aMenuItems is not None:
            if not isinstance(aMenuItems, list):
                if not isinstance(aMenuItems, tuple):
                    # Just straight menu item
                    aMenuItems = [('Plugins', aMenuItems)]
                else:
                    # Wrap tuple in a list
                    aMenuItems = [aMenuItems]
            for sMenu, oMenuItem in aMenuItems:
                if sMenu in dAllMenus:
                    dAllMenus[sMenu].add(oMenuItem)
                else:
                    # Plugins acts as a catchall Menu
                    oCatchAllMenu.add(oMenuItem)

    # pylint: disable-msg=R0201
    # We expect children to override these when needed
    def get_menu_item(self):
        """Return a list of ('Menu', gtk.MenuItems) pairs for the plugin or
           None if no menu item is needed."""
        return None

    def setup(self):
        """Handle any setup needed for the plugin after the main window has
           been initialised.

           Currently used to prompt for downloads, etc.
           """
        return None

    def get_toolbar_widget(self):
        """Return an arbitary gtk.Widget which is added to a VBox between the
           menu and the scrolled display area.

           Return None is no toolbar Widget is needed
           """
        return None

    # pylint: disable-msg=W0613
    # sType can be used by children of this class
    def get_frame_from_config(self, sType):
        """Hook for plugins which supply a frame in the Main window.

           Allows them to restore from the config file properly.
           """
        return None

    # pylint: enable-msg=R0201

    # Utility Functions / Plugin API

    def check_model_type(self):
        """Check whether the plugin should register on this frame."""
        if self._cModelType in self.aModelsSupported:
            return True
        return False

    def check_versions(self):
        """Check whether the plugin supports the current version of
           the Sutekh database tables."""
        oDBVer = DatabaseVersion()
        for oTable, aVersions in self.dTableVersions.iteritems():
            if not oDBVer.check_table_in_versions(oTable, aVersions):
                return False
        # If nothing is specified, currently we assume everything is A-OK
        return True

    def open_cs(self, sPCS):
        """Open a physical card set in the GUI."""
        self.parent.add_new_physical_card_set(sPCS)

    def reload_pcs_list(self):
        """Refresh the physical card set list if it is visible."""
        self.parent.reload_pcs_list()

    def reload_all(self):
        """Reload all views."""
        self.parent.reload_all()

    def get_card_set(self):
        """Get the Card Set for this view."""
        oCardSet = None
        if self._cModelType is PhysicalCardSet:
            # pylint: disable-msg=E1101
            # sqlobject confuses pylint
            oCardSet = PhysicalCardSet.byName(self.view.sSetName)
        return oCardSet

    def get_all_cards(self):
        """Get the cards from the card set."""
        if self._cModelType is PhysicalCardSet:
            return self.model.get_card_iterator(None)
        return []




# test_WriteCSV.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Test Writing a card set to an CSV file"""

from sutekh.tests.TestCore import SutekhTest
from sutekh.tests.core.test_PhysicalCardSet import aCardSetNames, \
        get_phys_cards
from sutekh.core.SutekhObjects import PhysicalCardSet
from sutekh.io.WriteCSV import  WriteCSV
import unittest

sExpected1 = """"Card Name", "Expansion", "Number"
".44 Magnum", "Jyhad", 1
".44 Magnum", "Unknown Expansion", 3
"AK-47", "Lords of the Night", 1
"AK-47", "Unknown Expansion", 1
"Abbot", "Third Edition", 1
"Abbot", "Unknown Expansion", 1
"Abebe", "Unknown Expansion", 1
"Abombwe", "Legacy of Blood", 1
"Abombwe", "Unknown Expansion", 1
"Alan Sovereign (Advanced)", "Promo-20051001", 1
"The Path of Blood", "Lords of the Night", 1
"""

sExpected2 = """".44 Magnum", "Jyhad", 1
".44 Magnum", "Unknown Expansion", 3
"AK-47", "Lords of the Night", 1
"AK-47", "Unknown Expansion", 1
"Abbot", "Third Edition", 1
"Abbot", "Unknown Expansion", 1
"Abebe", "Unknown Expansion", 1
"Abombwe", "Legacy of Blood", 1
"Abombwe", "Unknown Expansion", 1
"Alan Sovereign (Advanced)", "Promo-20051001", 1
"The Path of Blood", "Lords of the Night", 1
"""


sExpected3 = """"Card Name", "Number"
".44 Magnum", 4
"AK-47", 2
"Abbot", 2
"Abebe", 1
"Abombwe", 2
"Alan Sovereign (Advanced)", 1
"The Path of Blood", 1
"""

sExpected4 = """".44 Magnum", 4
"AK-47", 2
"Abbot", 2
"Abebe", 1
"Abombwe", 2
"Alan Sovereign (Advanced)", 1
"The Path of Blood", 1
"""


class CSVWriterTests(SutekhTest):
    """class for the CSV deck writer tests"""

    def test_deck_writer(self):
        """Test HTML deck writing"""
        # pylint: disable-msg=E1101, R0915, R0914
        # E1101: SQLObject + PyProtocols magic confuses pylint
        # R0915, R0914: Want a long, sequential test case to minimise
        # repeated setups, so it has lots of lines + variables
        aAddedPhysCards = get_phys_cards()
        # We have a physical card list, so create some physical card sets
        oPhysCardSet1 = PhysicalCardSet(name=aCardSetNames[0])
        oPhysCardSet1.comment = 'A test comment'
        oPhysCardSet1.author = 'A test author'

        for oCard in aAddedPhysCards:
            oPhysCardSet1.addPhysicalCard(oCard.id)
            oPhysCardSet1.syncUpdate()
        oPhysCardSet1.addPhysicalCard(aAddedPhysCards[0])
        oPhysCardSet1.addPhysicalCard(aAddedPhysCards[0])
        oPhysCardSet1.syncUpdate()

        # Check output

        oWriter = WriteCSV(True, True)
        sTempFileName =  self._create_tmp_file()
        fOut = file(sTempFileName, 'w')
        oWriter.write(fOut, oPhysCardSet1)
        fOut.close()

        fIn = open(sTempFileName, 'rU')
        sData = fIn.read()
        fIn.close()

        self.assertEqual(sData, sExpected1)

        oWriter = WriteCSV(False, True)
        sTempFileName =  self._create_tmp_file()
        fOut = file(sTempFileName, 'w')
        oWriter.write(fOut, oPhysCardSet1)
        fOut.close()

        fIn = open(sTempFileName, 'rU')
        sData = fIn.read()
        fIn.close()

        self.assertEqual(sData, sExpected2)

        oWriter = WriteCSV(True, False)
        sTempFileName =  self._create_tmp_file()
        fOut = file(sTempFileName, 'w')
        oWriter.write(fOut, oPhysCardSet1)
        fOut.close()

        fIn = open(sTempFileName, 'rU')
        sData = fIn.read()
        fIn.close()

        self.assertEqual(sData, sExpected3)

        oWriter = WriteCSV(False, False)
        sTempFileName =  self._create_tmp_file()
        fOut = file(sTempFileName, 'w')
        oWriter.write(fOut, oPhysCardSet1)
        fOut.close()

        fIn = open(sTempFileName, 'rU')
        sData = fIn.read()
        fIn.close()

        self.assertEqual(sData, sExpected4)




if __name__ == "__main__":
    unittest.main()



# test_ELDBDeckFileParser.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Test reading a card set from an ELDB deck file"""

import unittest
from sutekh.tests.TestCore import SutekhTest, DummyHolder
from sutekh.io.ELDBDeckFileParser import ELDBDeckFileParser

# Needs to be a SutekhTestCase so the name mapping cache test works
class TestELDBDeckFileParser(SutekhTest):
    """class for the ELDB deck file parser tests"""

    sTestText1 = """
        "Test Deck"
        "Anon Y Mous"
        "Simple test deck.

        http://www.example.url/in/description"
        4
        "Test Vamp 1"
        "Test Vamp 1"
        "Lazar Dobrescu"
        "Alan Sovereign (ADV)"
        18
        "Test Card 1"
        "Test Card 1"
        "Test Card 2"
        "Test Card 2"
        "Test Card 2"
        "Test Card 2"
        "Test Card 3"
        "Test Card 3"
        "Test Card 3"
        "Test Card 3"
        "Test Card 3"
        "Test Card 3"
        "Test Card 3"
        "Test Card 3"
        "Test Card 3"
        "Test Card 3"
        "Test Card 3"
        "Test Card 3"
        "Test Card 4"
        "Path of Blood, The"
        """

    def test_basic(self):
        """Run the input test."""
        oHolder = DummyHolder()
        oParser = ELDBDeckFileParser(oHolder)

        for sLine in self.sTestText1.split("\n"):
            oParser.feed(sLine + "\n")

        self.assertEqual(oHolder.name, "Test Deck")
        self.assertEqual(oHolder.author, "Anon Y Mous")
        self.failUnless(oHolder.comment.startswith(
            "Simple test deck."))
        self.failUnless(oHolder.comment.endswith("in/description"))

        aCards = oHolder.get_cards()

        self.assertEqual(len(aCards), 8)
        self.failUnless(("Test Vamp 1", 2) in aCards)
        self.failUnless((u"L\xe1z\xe1r Dobrescu", 1) in aCards)
        self.failUnless(("Test Card 1", 2 ) in aCards)
        self.failUnless(("Test Card 2", 4) in aCards)
        self.failUnless(("Test Card 3", 12) in aCards)
        self.failUnless(("Test Card 4", 1) in aCards)
        self.failUnless(("Alan Sovereign (Advanced)", 1) in aCards)
        self.failUnless(("The Path of Blood", 1) in aCards)

if __name__ == "__main__":
    unittest.main()



# CardSetCompare.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2006 Simon Cross <hodgestar@gmail.com>,
# Copyright 2006, 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details
"""Compare the contents of two card sets"""

import gtk
from sutekh.core.SutekhObjects import PhysicalCard, IPhysicalCard, \
        PhysicalCardSet, IAbstractCard, IExpansion
from sutekh.core.Filters import PhysicalCardSetFilter
from sutekh.gui.PluginManager import CardListPlugin
from sutekh.gui.SutekhDialog import SutekhDialog, do_complaint_error
from sutekh.gui.ScrolledList import ScrolledList
from sutekh.gui.AutoScrolledWindow import AutoScrolledWindow
from sutekh.gui.CreateCardSetDialog import CreateCardSetDialog
from sutekh.gui.CardSetManagementController import update_card_set

def _get_card_set_list(aCardSetNames, bUseExpansions):
    """Get the differences and common cards for the card sets."""
    # Only compare abstract cards
    dFullCardList = {}
    for sCardSetName in aCardSetNames:
        oFilter = PhysicalCardSetFilter(sCardSetName)
        oCardSet = oFilter.select(PhysicalCard)
        for oCard in oCardSet:
            # pylint: disable-msg=E1101
            # pylint doesn't see IAbstractCard methods
            oAbsCard = IAbstractCard(oCard)
            if bUseExpansions:
                if oCard.expansion:
                    tKey = (oAbsCard.name, oCard.expansion.name)
                else:
                    tKey = (oAbsCard.name, 'Unspecified Expansion')
            else:
                tKey = (oAbsCard.name, None)
            dFullCardList.setdefault(tKey, {aCardSetNames[0] : 0,
                aCardSetNames[1] : 0})
            dFullCardList[tKey][sCardSetName] += 1
    dDifferences = { aCardSetNames[0] : [], aCardSetNames[1] : [] }
    aCommon = []
    for tCardInfo in dFullCardList:
        iDiff = dFullCardList[tCardInfo][aCardSetNames[0]] - \
                dFullCardList[tCardInfo][aCardSetNames[1]]
        iCommon = min(dFullCardList[tCardInfo][aCardSetNames[0]],
                dFullCardList[tCardInfo][aCardSetNames[1]])
        if iDiff > 0:
            dDifferences[aCardSetNames[0]].append((tCardInfo[0], tCardInfo[1],
                iDiff))
        elif iDiff < 0:
            dDifferences[aCardSetNames[1]].append((tCardInfo[0], tCardInfo[1],
                abs(iDiff)))
        if iCommon > 0:
            aCommon.append((tCardInfo[0], tCardInfo[1], iCommon))
    return (dDifferences, aCommon)


class CardSetCompare(CardListPlugin):
    """Compare Two Card Sets

       Display a gtk.Notebook containing tabs for common cards, and cards
       only in each of the card sets.
       """
    dTableVersions = {PhysicalCardSet : [5]}
    aModelsSupported = [PhysicalCardSet]

    def get_menu_item(self):
        """Register on the 'Plugins' menu."""
        if not self.check_versions() or not self.check_model_type():
            return None
        oCompare = gtk.MenuItem("Compare with another Card Set")
        oCompare.connect("activate", self.activate)
        return ('Plugins', oCompare)

    # pylint: disable-msg=W0613
    # oWidget required by gtk function signature
    def activate(self, oWidget):
        """Create the dialog for choosing the second card set."""
        oDlg = SutekhDialog("Choose Card Set to Compare with", self.parent,
                gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                (gtk.STOCK_OK, gtk.RESPONSE_OK,
                    gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
        oSelect = PhysicalCardSet.select().orderBy('name')
        oCSList = ScrolledList('Card Sets')
        oCSList.set_select_single()
        # pylint: disable-msg=E1101
        # pylint misses vbox methods
        oDlg.vbox.pack_start(oCSList)
        oUseExpansions = gtk.CheckButton("Consider Expansions as well")
        oDlg.vbox.pack_start(oUseExpansions)
        oCSList.set_size_request(150, 300)
        aVals = [oCS.name for oCS in oSelect if oCS.name != self.view.sSetName]
        oCSList.fill_list(aVals)
        oDlg.connect("response", self.handle_response, oCSList, oUseExpansions)
        oDlg.show_all()
        oDlg.run()
        oDlg.destroy()

    def handle_response(self, oWidget, iResponse, oCSList, oUseExpansions):
        """Handle response from the dialog."""
        if iResponse ==  gtk.RESPONSE_OK:
            aCardSetNames = [self.view.sSetName]
            aCardSetNames.extend(oCSList.get_selection())
            if len(aCardSetNames) != 2:
                do_complaint_error("Need to choose a card set to compare to")
            else:
                self.comp_card_sets(aCardSetNames, oUseExpansions.get_active())

    def comp_card_sets(self, aCardSetNames, bUseExpansions):
        """Display the results of comparing the card sets."""
        def format_list(aList, sColor):
            """Format the list of cards for display."""
            oLabel = gtk.Label()
            oAlign = gtk.Alignment()
            oAlign.add(oLabel)
            oAlign.set_padding(0, 0, 5, 0) # offset a little from the left edge
            if len(aList) > 0:
                sContents = ""
                aList.sort()
                for sCardName, sExpansion, iCount in aList:
                    sContents += '%(num)d X <span foreground = "%(color)s">' \
                            '%(name)s' % {
                                    'num' : iCount,
                                    'color' : sColor,
                                    'name' : sCardName,
                                    }
                    if sExpansion:
                        sContents += " (%s)</span>\n" % sExpansion
                    else:
                        sContents += "</span>\n"
                oLabel.set_markup(sContents)
            else:
                oLabel.set_text('No Cards')
            return oAlign

        def make_page(oList, aCardData):
            """Setup the list + button for the notebook"""
            oPage = gtk.VBox(False)
            oPage.pack_start(AutoScrolledWindow(oList, True), True)
            if aCardData:
                oButton = gtk.Button('Create Card Set from this list')
                oButton.connect('clicked', self.create_card_set,
                        aCardData)
                oPage.pack_start(oButton, False)
            return oPage

        (dDifferences, aCommon) = _get_card_set_list(aCardSetNames,
                bUseExpansions)
        oResultDlg = SutekhDialog("Card Comparison", self.parent,
                gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                (gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE))
        oNotebook = gtk.Notebook()
        oNotebook.set_scrollable(True)
        oNotebook.popup_enable()
        oHeading = gtk.Label()
        sTabText = 'Common Cards'
        oHeading.set_markup('<span foreground = "blue">%s</span>' % sTabText)
        oComm = format_list(aCommon, 'green')
        oPage = make_page(oComm, aCommon)
        oNotebook.append_page_menu(oPage, oHeading, gtk.Label(sTabText))
        oHeading = gtk.Label()
        sTabText ='Cards only in %s' % aCardSetNames[0]
        oHeading.set_markup('<span foreground = "red">%s</span>' % sTabText)
        oDiff1 = format_list(dDifferences[aCardSetNames[0]], 'red')
        oPage = make_page(oDiff1, dDifferences[aCardSetNames[0]])
        oNotebook.append_page_menu(oPage, oHeading, gtk.Label(sTabText))
        sTabText ='Cards only in %s' % aCardSetNames[1]
        oHeading = gtk.Label()
        oHeading.set_markup('<span foreground = "red">%s</span>' % sTabText)
        oDiff2 = format_list(dDifferences[aCardSetNames[1]], 'red')
        oPage = make_page(oDiff2, dDifferences[aCardSetNames[1]])
        oNotebook.append_page_menu(oPage, oHeading, gtk.Label(sTabText))
        # pylint: disable-msg=E1101
        # pylint misses vbox methods
        oResultDlg.vbox.pack_start(oNotebook)
        oResultDlg.set_size_request(600, 600)
        oResultDlg.show_all()
        oResultDlg.run()
        oResultDlg.destroy()

    def create_card_set(self, oButton, aCardData):
        """Create a card set from the card list"""
        oDlg = CreateCardSetDialog(self.parent)
        oDlg.run()
        sCSName = oDlg.get_name()
        if not sCSName:
            return
        if PhysicalCardSet.selectBy(name=sCSName).count() != 0:
            do_complaint_error("Card Set %s already exists."
                    % sCSName)
            return
        oCardSet = PhysicalCardSet(name=sCSName)
        for sCardName, sExpansionName, iCnt in aCardData:
            if sExpansionName and sExpansionName != 'Unspecified Expansion':
                oExpansion = IExpansion(sExpansionName)
            else:
                oExpansion = None
            oCard = IAbstractCard(sCardName)
            oPhysCard = IPhysicalCard((oCard, oExpansion))
            # pylint: disable-msg=W0612, E1101
            # W0612 - iNum is just loop counter
            # E1101 - sqlobject confuses pylint
            for iNum in range(iCnt):
                oCardSet.addPhysicalCard(oPhysCard)
        update_card_set(oCardSet, oDlg, self.parent, None)
        self.open_cs(sCSName)


# pylint: disable-msg=C0103
# accept plugin name
plugin = CardSetCompare



# SutekhUtility.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2006 Simon Cross <hodgestar@gmail.com>
# Copyright 2006 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details
# Misc Useful functions needed in several places. Mainly to do with database
# management. Seperated out from SutekhCli and other places, NM, 2006

"""Misc functions needed in various places in Sutekh."""

from sutekh.core.SutekhObjects import VersionTable, flush_cache
from sutekh.core.DatabaseVersion import DatabaseVersion
from sutekh.io.WhiteWolfParser import WhiteWolfParser
from sutekh.io.RulingParser import RulingParser
from sqlobject import sqlhub
import tempfile, os, sys

def refresh_tables(aTables, oConn):
    """Drop and recreate the given list of tables"""
    aTables.reverse()
    for cCls in aTables:
        cCls.dropTable(ifExists=True, connection=oConn)
    aTables.reverse()
    oVerHandler = DatabaseVersion(oConn)
    # Make sure we recreate the database version table
    oVerHandler.expire_table_conn(oConn)
    oVerHandler.ensure_table_exists(oConn)
    if not oVerHandler.set_version(VersionTable, VersionTable.tableversion,
            oConn):
        return False
    for cCls in aTables:
        cCls.createTable(connection=oConn)
        if not oVerHandler.set_version(cCls, cCls.tableversion, oConn):
            return False
    flush_cache()
    return True

def read_white_wolf_list(aWwFiles, oLogHandler=None):
    """Parse in a new White Wolf cardlist

       aWwList is a list of objects with a .open() method (e.g.
       sutekh.io.WwFile.WwFile's)
       """
    flush_cache()
    oOldConn = sqlhub.processConnection
    sqlhub.processConnection = oOldConn.transaction()
    oParser = WhiteWolfParser(oLogHandler)
    for oFile in aWwFiles:
        fIn = oFile.open()
        for sLine in fIn:
            oParser.feed(sLine)
        fIn.close()
    sqlhub.processConnection.commit(close=True)
    sqlhub.processConnection = oOldConn

def read_rulings(oRulings, oLogHandler=None):
    """Parse a new White Wolf rulings file

       oRulings is an object with a .open() method (e.g. a sutekh.io.WwFile.WwFile)
       """
    flush_cache()
    oOldConn = sqlhub.processConnection
    sqlhub.processConnection = oOldConn.transaction()
    oParser = RulingParser(oLogHandler)
    fIn = oRulings.open()
    for sLine in fIn:
        oParser.feed(sLine)
    fIn.close()
    sqlhub.processConnection.commit(close=True)
    sqlhub.processConnection = oOldConn

def gen_temp_file(sBaseName, sDir):
    """Simple wrapper around tempfile creation - generates the name and closes
       the file
       """
    (fTemp, sFilename) = tempfile.mkstemp('.xml', sBaseName, sDir)
    # This may not be nessecary, but the available documentation
    # suggests that, on Windows NT anyway, leaving the file open will
    # cause problems if we try to reopen it
    os.close(fTemp)
    # There is a race condition here, but since Sutekh should not be running
    # with elevated priveleges, this should never be a security issues
    # The race requires something to delete and replace the tempfile,
    # I don't see it being triggered accidently
    return sFilename

def gen_temp_dir():
    """Create a temporary directory using mkdtemp"""
    sTempdir = tempfile.mkdtemp('dir', 'sutekh')
    return sTempdir

def safe_filename(sFilename):
    """Replace potentially dangerous and annoying characters in the name -
       used to automatically generate sensible filenames from card set names
       """
    sSafeName = sFilename
    sSafeName = sSafeName.replace(" ", "_") # I dislike spaces in filenames
    # Prevented unexpected filesystem issues
    sSafeName = sSafeName.replace("/", "_")
    sSafeName = sSafeName.replace("\\", "_") # ditto for windows
    return sSafeName

def prefs_dir(sApp):
    """Return a suitable directory for storing preferences and other
       application data."""
    if sys.platform.startswith("win") and "APPDATA" in os.environ:
        return os.path.join(os.environ["APPDATA"], sApp)
    else:
        return os.path.join(os.path.expanduser("~"), ".%s" % sApp.lower())

def ensure_dir_exists(sDir):
    """Check that a directory exists and create it if it doesn't.
       """
    if os.path.exists(sDir):
        assert os.path.isdir(sDir)
    else:
        os.makedirs(sDir)

def sqlite_uri(sPath):
    """Create an SQLite db URI from the path to the db file.
       """
    sDbFile = sPath.replace(os.sep, "/")

    sDrive, sRest = os.path.splitdrive(sDbFile)
    if sDrive:
        sDbFile = "/" + sDrive.rstrip(':') + "|" + sRest
    else:
        sDbFile = sRest

    return "sqlite://" + sDbFile

def pretty_xml(oElement, iIndentLevel=0):
    """
    Helper function to add whitespace text attributes to a ElementTree.
    Makes for 'pretty' indented XML output.
    Based on the example indent function at http://effbot.org/zone/element-lib.htm [22/01/2008]
    """
    sIndent = "\n" + iIndentLevel * "  "
    if len(oElement):
        if not oElement.text or not oElement.text.strip():
            oElement.text = sIndent + "  "
            for oSubElement in oElement:
                pretty_xml(oSubElement, iIndentLevel + 1)
            # Reset indentation level for last child element
            # pylint: disable-msg=W0631
            # We know SubElement will exist because of the len check above
            if not oSubElement.tail or not oSubElement.tail.strip():
                oSubElement.tail = sIndent
    else:
        if iIndentLevel and (not oElement.tail or not oElement.tail.strip()):
            oElement.tail = sIndent





# FullBackup.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2007 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details
"""Plugin to wrap zipfile backup and restore methods"""

from sutekh.gui.PluginManager import CardListPlugin
from sutekh.gui.SutekhDialog import do_complaint_error, do_complaint_warning
from sutekh.gui.SutekhFileWidget import SutekhFileDialog
from sutekh.gui.ProgressDialog import ProgressDialog, SutekhCountLogHandler
from sutekh.io.ZipFileWrapper import ZipFileWrapper
import gtk
import os

class FullBackup(CardListPlugin):
    """Provide access to ZipFileWrapper's backup and restore methods.

       Handle GUI aspects associated with restoring (ensuring everything
       reloads, etc.)
       """

    dTableVersions = {}
    aModelsSupported = ["MainWindow"]

    # Dialog and Menu Item Creation

    def get_menu_item(self):
        """Register on the Plugins menu"""
        if not self.check_versions() or not self.check_model_type():
            return None

        oMenuItem = gtk.MenuItem("Backup")
        oMenu = gtk.Menu()
        oMenuItem.set_submenu(oMenu)

        oBackup = gtk.MenuItem("Save a Full Backup")
        oBackup.connect("activate", self.activate_backup)
        oRestore = gtk.MenuItem("Restore a Full Backup")
        oRestore.connect("activate", self.activate_restore)

        oMenu.add(oBackup)
        oMenu.add(oRestore)

        return ('Plugins', oMenuItem)

    # Menu responses

    # pylint: disable-msg=W0613
    # oWidget needed by gtk function signature
    def activate_backup(self, oWidget):
        """Handle backup request"""
        oDlg = self.make_backup_dialog()
        oDlg.run()

    # oWidget needed by gtk function signature
    def activate_restore(self, oWidget):
        """Handle restore request"""
        oDlg = self.make_restore_dialog()
        oDlg.run()
    # pylint: enable-msg=W0613

    # Backup

    def make_backup_dialog(self):
        """Create file dialog for backup"""
        sName = "Choose a file to save the full backup to ..."

        oDlg = SutekhFileDialog(self.parent, sName,
                oAction=gtk.FILE_CHOOSER_ACTION_SAVE,
                oButtons=(gtk.STOCK_OK, gtk.RESPONSE_OK,
                    gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
        oDlg.set_name("Sutekh.dialog")
        oDlg.add_filter_with_pattern('Zip Files', ['*.zip', '*.ZIP'])
        oDlg.set_do_overwrite_confirmation(True)

        oDlg.connect("response", self.handle_backup_response)
        oDlg.set_local_only(True)
        oDlg.set_select_multiple(False)
        oDlg.show_all()

        return oDlg

    # pylint: disable-msg=R0201
    # This could be a function, but that won't add any clarity to this code
    def handle_backup_response(self, oDlg, oResponse):
        """Handle response from backup dialog"""
        if oResponse == gtk.RESPONSE_OK:
            sFile = oDlg.get_filename()
            oDlg.destroy()
            # pylint: disable-msg=W0703
            # we really do want all the exceptions
            try:
                oLogHandler = SutekhCountLogHandler()
                oProgressDialog = ProgressDialog()
                oProgressDialog.set_description("Saving backup")
                oLogHandler.set_dialog(oProgressDialog)
                oProgressDialog.show()
                oFile = ZipFileWrapper(sFile)
                oFile.do_dump_all_to_zip(oLogHandler)
                oProgressDialog.destroy()
            except Exception, oException:
                oProgressDialog.destroy()
                sMsg = "Failed to write backup.\n\n%s" % oException
                do_complaint_error(sMsg)
        else:
            oDlg.destroy()

    # pylint: enable-msg=R0201

    # Restore

    def make_restore_dialog(self):
        """Create file chooser dialog for restore"""
        sName = "Restore a Full Backup ...."

        oDlg = SutekhFileDialog(self.parent, sName,
                oAction=gtk.FILE_CHOOSER_ACTION_OPEN,
                oButtons=(gtk.STOCK_OK, gtk.RESPONSE_OK,
                    gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
        oDlg.set_name("Sutekh.dialog")
        oDlg.add_filter_with_pattern('Zip Files', ['*.zip', '*.ZIP'])

        oWarning = gtk.Label("This will delete all existing Card Sets")
        # pylint: disable-msg=E1101
        # plint doesn't pick up vbox methods correctly
        oDlg.vbox.pack_start(oWarning, expand=False)
        oDlg.vbox.reorder_child(oWarning, 0)
        oDlg.connect("response", self.handle_restore_response)
        oDlg.set_local_only(True)
        oDlg.set_select_multiple(False)
        oDlg.show_all()

        return oDlg

    def handle_restore_response(self, oDlg, oResponse):
        """Handle response from the restore dialog"""
        if oResponse == gtk.RESPONSE_OK:
            sFile = oDlg.get_filename()
            oDlg.destroy()
            bContinue = True

            if not os.path.exists(sFile):
                bContinue = do_complaint_warning(
                        "Backup file %s does not seem to exist."
                        % sFile) != gtk.RESPONSE_CANCEL

            if bContinue:
                try:
                    aEditable = self.parent.get_editable_panes()
                    oLogHandler = SutekhCountLogHandler()
                    oProgressDialog = ProgressDialog()
                    oProgressDialog.set_description("Restoring backup")
                    oLogHandler.set_dialog(oProgressDialog)
                    oProgressDialog.show()
                    oFile = ZipFileWrapper(sFile)
                    oFile.do_restore_from_zip(self.cardlookup, oLogHandler)
                    # restore successful, refresh display
                    aMessages = oFile.get_warnings()
                    if aMessages:
                        sMsg = "The following warnings were reported:\n%s" % \
                                "\n".join(aMessages)
                        do_complaint_warning(sMsg)
                    # Id's will not be preserved
                    self.parent.update_to_new_db()
                    oProgressDialog.destroy()
                    self.parent.restore_editable_panes(aEditable)
                # pylint: disable-msg=W0703
                # we really do want all the exceptions
                except Exception, oException:
                    oProgressDialog.destroy()
                    sMsg = "Failed to restore backup.\n\n%s" % oException
                    do_complaint_error(sMsg)
        else:
            oDlg.destroy()


# pylint: disable-msg=C0103
# accept plugin name
plugin = FullBackup



# CardSetFrame.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Frame holding a CardSet
# Copyright 2006, 2007 Neil Muller <drnlmuller+sutekh@gmail.com>
# Copyright 2006 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""Sutekh Frame for holding Card Sets"""

from sqlobject import SQLObjectNotFound
from sutekh.core.SutekhObjects import PhysicalCardSet, \
        IPhysicalCardSet
from sutekh.gui.CardListFrame import CardListFrame
from sutekh.gui.CardSetMenu import CardSetMenu
from sutekh.gui.CardSetController import CardSetController

class CardSetFrame(CardListFrame, object):
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    """class for Card Set frames.

       Handles most of the functionality - subclasses set the style name
       and the various other properties correctly for the type.
       """
    _cModelType = PhysicalCardSet

    def __init__(self, oMainWindow, sName, tInfo=None):
        super(CardSetFrame, self).__init__(oMainWindow)
        try:
            # pylint: disable-msg=W0612
            # oCS just used for existance test, so ignored
            oCS = IPhysicalCardSet(sName)
        except SQLObjectNotFound:
            raise RuntimeError("Card Set %s does not exist" % sName)
        self._oController = CardSetController(sName,
                oMainWindow, self)

        if tInfo:
            self.set_model_modes(tInfo)

        self.init_plugins()

        self._oMenu = CardSetMenu(self, self._oController, self._oMainWindow,
                sName)
        self.set_name("physical card set card list")
        self.add_parts()

        self.update_name(sName)

        self._oController.view.check_editable()

    # pylint: disable-msg=W0212
    # We allow access via these properties
    name = property(fget=lambda self: self._sName, doc="Frame Name")
    # pylint: enable-msg=W0212

    def cleanup(self):
        """Cleanup function called before pane is removed by the Main Window"""
        self._oMainWindow.reload_pcs_list()
        # Clean up the signal handlers, to avoid problems
        self._oController.cleanup()

    def update_name(self, sNewName):
        """Update the frame name to the current card set name."""
        # pylint: disable-msg=W0201
        # this is called from __init__, so OK
        self.sSetName = sNewName
        self._sName = sNewName
        self.set_title(self.sSetName)

    def update_to_new_db(self):
        """Re-associate internal data against the database.

           Needed for re-reading WW cardlists and such.
           Instruct controller + model to update themselves,
           then reload.
           """
        self._oController.update_to_new_db()
        self.reload()

    def get_model_modes(self):
        """Get the model modes"""
        return self._oController.model.iExtraLevelsMode, \
                self._oController.model.iParentCountMode, \
                self._oController.model.iShowCardMode

    def set_model_modes(self, tInfo):
        """Set the model + menu to the saved modes"""
        iExtraLevelsMode, iParentCountMode, iShowCardMode = tInfo
        self._oController.model.iExtraLevelsMode = iExtraLevelsMode
        self._oController.model.iParentCountMode = iParentCountMode
        self._oController.model.iShowCardMode = iShowCardMode

    def do_queued_reload(self):
        """Do a defferred reload if one was set earlier"""
        if self._bNeedReload:
            self.reload()
        self._bNeedReload = False

